// IMP Compiler for 8086 - pass 3

// This reads an intermediate object file produced by the
// second pass, performs the jump and stack allocation fixups,
// and writes a standard Microsoft format object file.  It
// reads the input file twice - once to collect all the jump
// and stack information, and then a second time to actually
// write the object file.
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "pass3.h"

static char path_buffer[_MAX_PATH];	// working areas for file lookup
static char drive[_MAX_DRIVE];
static char dir[_MAX_DIR];
static char fname[_MAX_FNAME];
static char ext[_MAX_EXT];

// Pass 3 builds an in-store model of the application as a series
// literal code blocks, data blocks, and so on.
struct item {
	int what;			// what this block describes
	int address;		// the address in the image
	int info;			// type dependent extra information
	int size;			// size this block occupies in the image (generally in bytes)
};

#define MAXITEM 10000
struct item m[MAXITEM];
int nm = 0;

// Jumps and calls all go to logical labels (a label ID is a 16 bit number)
// We collect a simple database of labels and their corresponding code addresses
struct label {
	int labelid;
	int address;
};
#define MAXLABEL 1500
struct label labels[MAXLABEL];
int nl = 1;

// The entry to each routine includes code to move the stack frame for
// local variables.  Pass 2 plants a fixup record at the end of the
// routine.  We use this table to match fixups with their corresponding
// entry sequence, and we also use it to plant trap records
struct stfix {
	int id;		// arbitrary ID passed by Pass 2 (actually derived from P2's nominal code address)
	int hint;	// pointer to the M record corresponding to the entry code
	int events;	// events trapped in this subroutine (a 16 bit bitmask)
	int start;	// actual start address of subroutine
	int trap;	// label of the event trap entry point (if events != 0)
	int evfrom;	// label of the start of the event protected area
	int end;	// actual end address of subroutine
	int namep;	// pointer to debug name of this routine
};
#define MAXSTACK 500
struct stfix stackfix[MAXSTACK];
int ns = 0;

// The name dictionary is filled by external import or export
// names, zero terminated - the M record points to the first
// character of the corresponding name
#define MAXNAME 5000
char named[MAXNAME];
int namedp = 0;

// Line number information is collected as the object file is
// written so that we can write a linenumber segment for the
// debugger
struct lineno {
	int line;
	int offset;
};

#define MAXLINENO 4000
struct lineno lines[MAXLINENO];
int nlines = 0;
int lastlinead = -1;	// the code address at the last line record assigned

// Pass 2 passes on all symbols that are "spec'ed", and allocates an index
// in sequence.  We check to see if they ever actually get used so that we
// can prune the unwanted ones, and remap the indexes into the symbol table
// accordingly...
struct symspec {
	int used;		// a boolean flag
	int p3index;	// the actual symbol index
};

#define MAXSPECS 250
struct symspec specs[MAXSPECS];
int nspecs = 0;

// Pass3 needs to know whether this is a main program or a
// file of external routines, because main programs get a
// special symbol defined for the trap table, and a special
// name for the table segment...
int mainprog = 0;

//////// Database routines

// report this line number as being at this address
static void newlineno(int line, int addr)
{
	// is this line the same address as we already have?
	if (addr == lastlinead)	// lines have advanced, but code didn't
	{
		lines[nlines-1].line = line;	// update current record
	}
	else
	{
		if (nlines < MAXLINENO)		// only if there is room
		{
			lines[nlines].line = line;
			lines[nlines].offset = addr;
			nlines += 1;
			lastlinead = addr;
		}
	}
}

// return the index of the next item block
static int newitem()
{
	if (nm == MAXITEM)
	{
		fprintf(stderr, "Program too big\n");
		exit(1);
	}
	nm = nm + 1;
	return (nm - 1);
}

// return the index of the next label record
static int newlabel()
{
	if (nl == MAXLABEL)
	{
		fprintf(stderr, "Too many labels\n");
		exit(1);
	}
	nl = nl + 1;
	return (nl - 1);
}

// return the index of the next procedure/stack record
static int newstack()
{
	if (ns == MAXSTACK)
	{
		fprintf(stderr, "Too many subroutines\n");
		exit(1);
	}
	ns = ns + 1;
	return (ns - 1);
}

// copy a new name into the name dictionary, and return
// the index of the first character
static int newname(char * name)
{
	int l;

	l = strlen(name);
	if ((l + namedp) >= MAXNAME)
	{
		fprintf(stderr, "Too many names\n");
		exit(1);
	}
	strcpy(&named[namedp], name);
	namedp += l + 1;
	return (namedp - (l + 1));
}

// The first pass through the input file, where we collect all the
// data we will need to map out the object code
static void readpass1(FILE *input)
{
	int type, length, current, ptr, id, value, cad;
	unsigned char buffer[256];

	cad = 0;

	current = newitem();
	m[current].what = IF_OBJ;
	m[current].size = 0;
	for(;;)
	{
		readifrecord(input, type, length, buffer);
		if (type < 0)	// end of file marker
			return;
		switch(type)
		{
		case  IF_REFEXT:	// external name relative offset code word
			// update our "used" flag
			id = (buffer[1]<<8)|buffer[0];
			specs[id].used = 1;
			// and fall through to treat as a general code word
		case  IF_DATA:	// dataseg offset word
		case  IF_CONST:	// const seg offset word
		case  IF_DISPLAY:	// display seg offset word
		case  IF_BSS:	// BSS segment offset code word
		case  IF_SWT:	// SWITCH table segment offset code word
			if (length != WORDSIZE)
				fprintf(stderr, "Oops - length screwup!\n");
			// fall through
		case  IF_OBJ:	// plain object code
			// All of these are treated as "object code" for the first passs
			if (m[current].what != IF_OBJ)
			{
				current = newitem();
				m[current].what = IF_OBJ;
				m[current].size = 0;
			}
			// we don't bother to remember the code, just how big it is..
			m[current].size += length;
			cad += length;
			break;

		case  IF_JUMP:	// unconditional jump to label
		case  IF_CALL:	// call a label
			current = newitem();
			m[current].what = type;
			m[current].size = 3;	// assume long to begin with
			m[current].info = buffer[0] | (buffer[1] << 8);	// target label number
			cad += 3;
			break;

		case  IF_JCOND:	// cond jump to label JE, JNE, JLE, JL, JGE, JG
			current = newitem();
			m[current].what = type;
			m[current].size = 5;	// assume long to begin with
			// condition code is buffer[0]
			m[current].info = buffer[1] | (buffer[2] << 8);	// target label number
			cad += 5;
			break;

		case  IF_LABEL:	// define a label
			current = newitem();
			m[current].what = type;
			m[current].size = 0;	// labels occupy no space
			m[current].info = buffer[0] | (buffer[1] << 8);	// label number
			break;

		case  IF_FIXUP:	// define location for stack fixup instruction
			current = newitem();
			m[current].what = type;
			m[current].size = 4;	// space will be an ENTER instruction (C8, nnnn, ll)
			m[current].info = 0;	// amount to subtract from the stack will be filled later
			cad += 4;
			ptr = newstack();
			stackfix[ptr].id = buffer[0] | (buffer[1] << 8);	// id number for fixup
			stackfix[ptr].hint = current;	// point to this code item
			stackfix[ptr].events = 0;		// assume no events trapped
			stackfix[ptr].trap = 0;			// no label
			buffer[length] = 0;				// make sure we zero terminate the proc name
			stackfix[ptr].namep = newname((char *)&buffer[3]);		// debug name
			break;

		case  IF_SETFIX:// stack fixup <location> <amount> <eventmask> <event entry>
			id = buffer[0] | (buffer[1] << 8);	// id number for fixup
			for (ptr = 0; ptr < ns; ptr++)
			{
				if (stackfix[ptr].id == id)		// found it
				{
					id = stackfix[ptr].hint;	// point to M record
					value = buffer[2] | (buffer[3] << 8);	// amount to subtract
					// compiler passes value as a 16 bit negative number, but we're going
					// to plant an ENTER instruction, so we make it positive...
					value = - value;
					value &= 0xffff;
					m[id].info = value;
					// now fill in the event stuff...
					stackfix[ptr].events = buffer[4] | (buffer[5] << 8);
					stackfix[ptr].trap = buffer[6] | (buffer[7] << 8);
					stackfix[ptr].evfrom = buffer[8] | (buffer[9] << 8);
					break;
				}
			}
			if (ptr == ns)
				fprintf(stderr, "Stack fixup for undefined ID?\n");
			break;

		case IF_REQEXT:	// external name spec
			// clear it and count it...
			nspecs += 1;
			if (nspecs == MAXSPECS)
			{
				fprintf(stderr, "Too many %spec's\n");
				exit(1);
			}
			specs[nspecs].used = 0;
			// fall through
		case IF_DEFEXTCODE:	// define a code label that is external
			current = newitem();
			m[current].what = type;
			m[current].size = 0;	// definitions/specs occupy no space
			buffer[length] = 0;		// make the name null terminated
			m[current].info = newname((char *)buffer);
			if (type == IF_DEFEXTCODE)
			{
				// this is a slightly cheesy way of finding if this is a main program
				if (strcmp((char *)buffer, "__impmain") == 0)
					mainprog = 1;	// This is a main program
			}
			break;


		case IF_DEFEXTDATA:	// define a data label that is external
			break;

		case IF_COTWORD:	// Constant table word
		case IF_DATWORD:	// Data segment word
		case IF_SWTWORD:	// switch table entry - actually a label ID
			if (m[current].what != type)
			{
				current = newitem();
				m[current].what = type;
				m[current].size = 0;
			}
			m[current].size += 2;	// another word
			break;

		case IF_SOURCE:	// name of the source file
		case IF_LINE:	// line number info for the debugger
			// do nothing - not even advance the "current"
			break;
		}
	}
}

// Given a label ID, return the index of the label record
// Return zero if there's no such label
static int findlabel(int id)
{
	int i;

	for (i = 1; i < nl; i++)
		if (labels[i].labelid == id)
			return i;

	return 0;
}

// Reset the label list and then do a pass through the data
// to set up label records.
static void initlabels()
{
	int cad, i, type, ptr, id;

	cad = 0;
	nl = 1;
	for (i = 0; i < nm; i++)
	{
		type = m[i].what;
		m[i].address = cad;
		if (type == IF_LABEL)
		{
			id = m[i].info;
			ptr = findlabel(id);		// Pass 2 redefines labels sometimes
			if (ptr == 0)
				ptr = newlabel();
			labels[ptr].labelid = id;
			labels[ptr].address = cad;
		}
		else
		{
			if ((type != IF_COTWORD) && (type != IF_DATWORD) && (type != IF_SWTWORD))
				cad += m[i].size;
		}
	}
}

// Simple routine that tries to "improve" the jumps.
// It returns "true" if it found an improvement.
// Unfortunately we need to iterate because every
// improvement moves all the downstream labels (which
// may allow further improvement), so this routine
// is called more than once.
static int improvejumpsizes()
{
	int i, type, ptr, distance, success;

	success = 0;
	for (i = 0; i < nm; i++)
	{
		type = m[i].what;
		if ((type == IF_JUMP) || (type == IF_JCOND))
		{
			if (m[i].size > 2)		// not already improved!
			{
				ptr = m[i].info;		// pick up label id
				ptr = findlabel(ptr);	// get table index
				distance = labels[ptr].address - (m[i].address + 2);
				if ((-127 < distance) && (distance < 127))	// could this be short?
				{
					m[i].size = 2;		//make it so
					success = 1;			//and tell the world we've done good
				}
			}
		}
	}
	return success;
}

// cond jump to label JE, JNE, JG, JGE, JL, JLE, JA, JAE, JB, JBE
static unsigned char jcondop[10] = {
	0x74, 0x75, 0x7F, 0x7D, 0x7C, 0x7E, 0x77, 0x73, 0x72, 0x76,
};

static unsigned char jfalseop[10] = {
	0x75, 0x74, 0x7E, 0x7C, 0x7D, 0x7F, 0x76, 0x72, 0x73, 0x77,
};

// Global counter used to plant linker segment size definitions
static int codeseg = 0;
static int dataseg = 0;
static int constseg = 0;
static int bssseg = 0;
static int displayseg = 32;
static int swtabseg = 0;
static int trapseg = 0;

// run through the database adding up the various segment sizes
void computesizes()
{
	int i, type, size;

	for (i = 0; i < nm; i++)
	{
		type = m[i].what;
		size = m[i].size;
		switch(type) {
		case  IF_OBJ:	// plain object code
		case  IF_DATA:	// dataseg offset word
		case  IF_CONST:	// const seg offset word
		case  IF_BSS:	// BSS seg offset word
		case IF_DISPLAY:	// display offset word
		case IF_SWT:	// SWITCH table segment offset code word
		case  IF_JUMP:	// unconditional jump to label
		case  IF_CALL:	// call a label
		case  IF_JCOND:	// cond jump to label JE, JNE, JLE, JL, JGE, JG
		case  IF_FIXUP:	// define location for stack fixup instruction
		case IF_REFEXT:	// external name relative offset code word
			codeseg += size;
			break;

		case IF_COTWORD:	// Constant table word
			constseg += size;
			break;

		case IF_DATWORD:	// Data segment word
			dataseg += size;
			break;

		case IF_SWTWORD:	// switch table entry - actually a label ID
			swtabseg += size;
			break;

		default:
			// all other directives don't consume space
			break;
		}
	}

	// finally, the trap segment will contain one record for
	// every procedure we've found
	trapseg = ns * 32;
}

static unsigned char codebuffer[128];
static int codeptr=0;

static void putcode(int b)
{
	codebuffer[codeptr] = (unsigned char)b;
	codeptr++;
}

static unsigned char outbuffer[128];
static int bufptr=0;

static void putbyte(int b)
{
	outbuffer[bufptr] = (unsigned char)b;
	bufptr++;
}

static void putw16(int w)
{
	putbyte(w & 255);
	putbyte(w >> 8);
}

// put an Intel style length delimited name in the buffer
static void putname(char *name)
{
	int i;

	i = strlen(name);
	putbyte(i);
	while(i--)
		putbyte(*name++);
}

// put a Microsoft style simple string in the buffer
static void putstring(char *name)
{
	while(*name != 0)
		putbyte(*name++);
}

static void putsegdef(FILE *output, int a, int c, int b, int p, int len, int seg, int cl, int ovl)
{
	bufptr = 0;
	putbyte((a << 5) | (c << 2) | (b << 1) | p);
	putw16(len);
	putbyte(seg);
	putbyte(cl);
	putbyte(ovl);
	writeobjectrecord(output, 0x98, bufptr, outbuffer);
	bufptr = 0;
}

// flush the pipeline in codebuffer into the output
static int flushdata(FILE *output, int seg, int offset)
{
	int i;

	if (codeptr == 0)
		return 0;

	bufptr = 0;
	putbyte(seg);
	putw16(offset);
	for(i = 0; i < codeptr; i++)
		putbyte(codebuffer[i]);
	codeptr = 0;

	writeobjectrecord(output, 0xA0, bufptr, outbuffer);
	bufptr = 0;
	return i;	// number of code bytes written
}

// write the exernal spec table to the symbol table area
static void putexternalspecs(FILE *output)
{
	int i, type, ptr, count, index;

	count = 0;
	bufptr = 0;
	index = 1;	// pass 2 spec's use 1-based IDs

	for (i = 0; i < nm; i++)
	{
		type = m[i].what;
		ptr = m[i].info;
		if (type == IF_REQEXT)
		{
			if (specs[index].used)
			{
				putname(&named[ptr]);
				putbyte(0);	// no type information
				count += 1;
				if (count > 3)	// flush every 4 entries
				{
					writeobjectrecord(output, 0x8C, bufptr, outbuffer);
					bufptr = 0;
					count = 0;
				}
			}
			index += 1;
		}
	}
	if (count != 0)
	{
		writeobjectrecord(output, 0x8C, bufptr, outbuffer);
		bufptr = 0;
	}
}

// plant the array of blocks used by %signal to trap
// events in the trap segment.  These blocks contain:
// <StartAddr16><EndAddr16><TrapAddr16><FromAddr16><EventMask16><Name[6]>
static void puttraptable(FILE *output)
{
	int i, j;
	int offset;
	struct stfix *sp;
	char *namep;

	offset = 0;

	for (i = 0; i < ns; i++)
	{
		sp = &stackfix[i];
		bufptr = 0;
		putbyte(7);			// Trap seg - note for both main prog and external, the trap seg is 7
		putw16(offset);
		putw16(sp->start);
		putw16(sp->end);
		// trap and evfrom are actually labels, so we look them up
		j = findlabel(sp->trap);
		putw16(labels[j].address);
		j = findlabel(sp->evfrom);
		putw16(labels[j].address);
		putw16(sp->events);
		namep = &named[sp->namep];
		for(j=0; j<6; j++)
		{
			putbyte(*namep);
			if (*namep) namep++;
		}
		writeobjectrecord(output, 0xA0, bufptr, outbuffer);
		// Of course the four code addresses we've just planted - start/end/entry/from - all
		// need to be relocated by the beginning of the code segment, so we do four
		// relocation records next...
		for (j=0; j < 4; j++)
		{
			bufptr = 0;
			// first plant the locat record.  Offset is 2 * j (for each address)
			putbyte(0x80 | (1 << 6) | (1 << 2) | (0 >> 8)); // M = 1, loc = 1
			putbyte(j * 2);	// second part of offset
			// now the fixdat field
			putbyte((0<<4)|(1<<2)|0);	// Frame is a segment (CODE), P = 1, target is seg and offset
			putbyte(1);			// index for CSEG
			putbyte(1);			// also index for CODE
			writeobjectrecord(output, 0x9C, bufptr, outbuffer);
		}
		offset += 16;
	}

	// if this was the main program file, we define a symbol that
	// corresponds to the base of the trap table, and one at the end.
	if (mainprog)
	{
		bufptr = 0;
		putbyte(1);	// DGROUP index
		putbyte(7);	// TRAP segment
		putname("__imptrapbase");
		putw16(0);	// offset 0
		putbyte(0);	// no TYPE information
		writeobjectrecord(output, 0x90, bufptr, outbuffer);

		// and again for the trap end
		bufptr = 0;
		putbyte(1);	// DGROUP index
		putbyte(9);	// TRAPEND segment
		putname("__imptrapend");
		putw16(0);	// offset 0
		putbyte(0);	// no TYPE information
		writeobjectrecord(output, 0x90, bufptr, outbuffer);
	}
}

// write the external definitions to the symbol table
static void putexternaldefs(FILE *output)
{
	int i, type, ptr;

	for (i = 0; i < nm; i++)
	{
		type = m[i].what;
		ptr = m[i].info;
		if (type == IF_DEFEXTCODE)
		{
			bufptr = 0;
			putbyte(0);	// no group index
			putbyte(1);	// CODE segment
			putname(&named[ptr]);
			putw16(m[i].address);
			putbyte(0);	// no TYPE information
			writeobjectrecord(output, 0x90, bufptr, outbuffer);
		}
	}
}

// Fill in the line number section of the object file
static void putlinenumbers(FILE *output)
{
	int i, line, addr, count;

	i = 0;
	count = 0;
	bufptr = 0;

	while (i < nlines)
	{
		if (count == 0)		// start a new block
		{
			putbyte(0);	// base group (ignored)
			putbyte(1);	// code segment
		}
		line = lines[i].line;
		addr = lines[i].offset;
		putw16(line);
		putw16(addr);
		i++;
		count++;
		if (count == 8)		// flush every eight
		{
			writeobjectrecord(output, 0x94, bufptr, outbuffer);
			bufptr = 0;
			count = 0;
		}
	}
	if (count != 0)		// flush remaining pipe
	{
		writeobjectrecord(output, 0x94, bufptr, outbuffer);
		bufptr = 0;
	}
}

// Main Pass - Reread the input file and write the object code
static void readwrite(FILE *input, FILE *output)
{
	int type, length, current, ptr, id, value, condition, cad, i, segidx;
	int cotp, datap, swtp;
	unsigned char buffer[128];

	current = 0;
	cad = 0;
	cotp = 0;
	datap = 0;
	swtp = 0;
	for(;;)
	{
		readifrecord(input, type, length, buffer);
		if (type < 0)
			break;
		switch(type)
		{
		case IF_SOURCE:	// name of the source file
			// HACK ALERT - we actually ignore the file name from PASS2 because
			// we've got a nicer library interface and we can get the REAL path
			putname((char *)path_buffer);
			writeobjectrecord(output, 0x80, bufptr, outbuffer);
			bufptr = 0;
			// this should now trigger us to output a bunch of other stuff
			putbyte(0);	// comment type
			putbyte(0xA1);	// codeview flags
			putbyte(1);		// version 1
			putbyte('C');
			putbyte('V');	// CodeView
			writeobjectrecord(output, 0x88, bufptr, outbuffer);
			bufptr = 0;
			putbyte(0);	// comment type
			putbyte(0);	// translator info
			putstring("NB IMP");
			writeobjectrecord(output, 0x88, bufptr, outbuffer);
			bufptr = 0;
			putbyte(0);	// comment type
			putbyte(0x9F);	// default library
			putstring("SLIBI77.LIB");	// small model IMP library
			writeobjectrecord(output, 0x88, bufptr, outbuffer);
			bufptr = 0;
			putbyte(0);	// comment type
			putbyte(0x9D);	//CPU model characters
			putstring("0s");	// 8086 small model
			writeobjectrecord(output, 0x88, bufptr, outbuffer);
			bufptr = 0;
			// Now publish our logical segment names
			putname("");		// 1
			putname("DGROUP");	// 2
			putname("_TEXT");	// 3
			putname("CODE");	// 4
			putname("_DATA");	// 5
			putname("DATA");	// 6
			putname("CONST");	// 7
			putname("_BSS");	// 8
			putname("BSS");		// 9
			putname("_DISPLAY");	// 10
			putname("_SWTAB");	// 11
			putname("_IMPTRA0");	// 12
			putname("_IMPTRA2");	// 13
			putname("_IMPTRA9");	// 14

			writeobjectrecord(output, 0x96, bufptr, outbuffer);
			// now associate the names with the logical segments
			// A = 2 -> relocatable word aligned
			// C = 2 -> public
			// B = 0 -> not big
			// P = 0 -> offsets are 16 bits
			putsegdef(output, 2, 2, 0, 0, codeseg, 3, 4, 1);
			putsegdef(output, 2, 2, 0, 0, dataseg, 5, 6, 1);
			putsegdef(output, 2, 2, 0, 0, constseg, 7, 7, 1);
			putsegdef(output, 2, 2, 0, 0, bssseg, 8, 9, 1);
			// C = 6 -> common - you can overlay all the displays
			putsegdef(output, 2, 6, 0, 0, displayseg, 10, 9, 1);
			putsegdef(output, 2, 2, 0, 0, swtabseg, 11, 7, 1);
			// main program defines all three possible trap segments to
			// get them in the right linker order
			if (mainprog)
			{
				putsegdef(output, 2, 2, 0, 0, trapseg, 12, 7, 1);	// trap base segment
				putsegdef(output, 2, 2, 0, 0, 0, 13, 7, 1);		// trap middle segment
				putsegdef(output, 2, 2, 0, 0, 16, 14, 7, 1);	// trap end segment
			}
			else
				putsegdef(output, 2, 2, 0, 0, trapseg, 13, 7, 1);
			// and finally group all the data segments together
			bufptr = 0;
			putbyte(2);	// DGROUP contains....
			putbyte(0xFF);
			putbyte(2);
			putbyte(0xFF);
			putbyte(3);
			putbyte(0xFF);
			putbyte(4);
			putbyte(0xFF);
			putbyte(5);
			putbyte(0xFF);
			putbyte(6);
			putbyte(0xFF);
			putbyte(7);
			if (mainprog)
			{
				putbyte(0xFF);
				putbyte(8);
				putbyte(0xFF);
				putbyte(9);
			}
			// all except the code segment
			writeobjectrecord(output, 0x9A, bufptr, outbuffer);
			bufptr = 0;
			putexternalspecs(output);
			putexternaldefs(output);
			break;

		case  IF_OBJ:	// plain object code
			if (m[current].what != IF_OBJ)
				current += 1;
			for (i = 0; i < length; i++)
				putcode(buffer[i]);
			if (codeptr > 90)
				cad += flushdata(output, 1, cad);
			break;

		case  IF_DATA:	// dataseg offset word
		case  IF_CONST:	// const seg offset word
		case IF_SWT:	// SWITCH table segment offset code word
			if (m[current].what != IF_OBJ)
				current += 1;
			for (i=0; i < WORDSIZE; i++)
				putcode(0);
			i = flushdata(output, 1, cad);	// fixup is at i - 2
			cad += i;
			i -= WORDSIZE;
			// first plant the locat record.
			bufptr = 0;
			putbyte(0x80 | (1 << 6) | (1 << 2) | (i >> 8)); // M = 1, loc = 1
			putbyte(i & 255);
			// now the fixdat field
			putbyte((1 << 4) | 0);	// Frame is a group (DGROUP), target is seg and offset
			putbyte(1);			// index for DGROUP
			if(type == IF_DATA)
				segidx = 2;
			if (type == IF_CONST)
				segidx = 3;
			if (type == IF_BSS)
				segidx = 4;
			if (type == IF_DISPLAY)
				segidx = 5;
			if (type == IF_SWT)
				segidx = 6;
			putbyte(segidx);
			for (i=0; i < WORDSIZE; i++)
				putbyte(buffer[i]);
			writeobjectrecord(output, 0x9C, bufptr, outbuffer);
			bufptr = 0;
			break;

		case  IF_REFEXT:	// external name relative offset code word
			if (m[current].what != IF_OBJ)
				current += 1;
			for (i=0; i < WORDSIZE; i++)
				putcode(0);		// plant some zeroes
			i = flushdata(output, 1, cad);	// fixup is at i - 2
			cad += i;
			i -= WORDSIZE;

			// reference index is in buffer[0,1]
			id = (buffer[1]<<8)|buffer[0];
			id = specs[id].p3index;			// remap according to our table

			// now plant the fixup record - the buffer bytes contain the external index
			// first the locator
			bufptr = 0;
			putbyte(0x80 | (0 << 6) | (1 << 2) | (i >> 8));	// M = 0, loctype = 1
			putbyte(i & 255);
			// now the fixdat field
			putbyte((5 << 4) | 6);	// Frame is copied from target, target is ext, offset 0
			if (id > 0x7F)			// is this a two byte index?
				putbyte(0x80 | (id >> 8));
			putbyte(id & 255);
			writeobjectrecord(output, 0x9C, bufptr, outbuffer);
			bufptr = 0;
			break;

		case  IF_JUMP:	// unconditional jump to label
			current += 1;
			id = buffer[0] | (buffer[1] << 8);	// target label number
			ptr = findlabel(id);
			value = labels[ptr].address;
			if (m[current].size == 2)			// short jump
			{
				putcode(0xEB);
				putcode(value - (m[current].address + 2));
			}
			else
			{
				putcode(0xE9);	// JMP
				value = value - (m[current].address + 3);
				putcode(value & 255);
				putcode(value >> 8);
			}
			break;

		case  IF_CALL:	// call a label
			current += 1;
			id = buffer[0] | (buffer[1] << 8);	// target label number
			ptr = findlabel(id);
			value = labels[ptr].address;
			putcode(0xE8);	// CALL
			value = value - (m[current].address + 3);
			putcode(value & 255);
			putcode(value >> 8);
			break;

		case  IF_JCOND:	// cond jump to label JE, JNE, JLE, JL, JGE, JG
			current += 1;
			condition = buffer[0];
			id = buffer[1] | (buffer[2] << 8);	// target label number
			ptr = findlabel(id);
			value = labels[ptr].address;
			if (m[current].size == 2)			// short jump
			{
				putcode(jcondop[condition]);
				putcode(value - (m[current].address + 2));
			}
			else
			{
				putcode(jfalseop[condition]);
				putcode(3);	// to skip the jmp
				putcode(0xE9);	// JMP
				value = value - (m[current].address + 5);
				putcode(value & 255);
				putcode(value >> 8);
			}
			break;

		case  IF_LABEL:	// define a label
			current += 1;
			break;

		case  IF_FIXUP:	// define location for stack fixup instruction
			current += 1;
			value = m[current].info;
			// For backward compatibility reasons (mostly because it kept messing
			// me up in development) we will plant suitable code whether this is
			// a "classic" 8086 fixup request - i.e. plant SUB SP,nnnn - or a new
			// style 80286 fixup request to plant ENTER nnnn,level.  We can tell
			// them apart because the classic passes only two parameter bytes.
			if (length == 2)
			{
				putcode(0x81);	// SUB
				putcode(0xEC);	// SP
				putcode(value & 255);	// Stack displacement
				putcode(value >> 8);
			}
			else
			{
				putcode(0xC8);	// ENTER
				putcode(value & 255);	// Stack displacement
				putcode(value >> 8);
				putcode(buffer[2]);		// level
			}
			// We now update our procedure record with the actual block start location
			id = buffer[0] | (buffer[1] << 8);	// id number for fixup
			for (ptr = 0; ptr < ns; ptr++)
			{
				if (stackfix[ptr].id == id)		// found it
				{
					stackfix[ptr].start = cad + codeptr;
					break;
				}
			}
			break;

		case  IF_SETFIX:// stack fixup <location> <amount> <eventmask> <event entry>
			// We don't need to do anything in the code stream here, but we use
			// this record to trigger an update of the end point in our block table
			id = buffer[0] | (buffer[1] << 8);	// id number for fixup
			for (ptr = 0; ptr < ns; ptr++)
			{
				if (stackfix[ptr].id == id)		// found it
				{
					stackfix[ptr].end = cad + codeptr;
					break;
				}
			}
			break;

		case IF_REQEXT:	// external name spec
		case IF_DEFEXTCODE:	// define a code label that is external
			// already taken care of, but need to advance "current"
			current += 1;
			break;

		case IF_DEFEXTDATA:	// define a data label that is external
			break;

		case IF_COTWORD:	// Constant table word
			if (m[current].what != IF_COTWORD)
			{
				current += 1;
				cad += flushdata(output, 1, cad);	// in case there's lurking code
			}
			bufptr = 0;
			putbyte(3);			// Const seg
			putw16(cotp);
			putbyte(buffer[0]);
			putbyte(buffer[1]);
			writeobjectrecord(output, 0xA0, bufptr, outbuffer);
			bufptr = 0;
			cotp += 2;
			break;

		case IF_DATWORD:	// Data segment word
			if (m[current].what != IF_DATWORD)
			{
				current += 1;
				cad += flushdata(output, 1, cad);	// in case there's lurking code
			}
			bufptr = 0;
			putbyte(2);			// Data seg
			putw16(datap);
			putbyte(buffer[0]);
			putbyte(buffer[1]);
			writeobjectrecord(output, 0xA0, bufptr, outbuffer);
			bufptr = 0;
			datap += 2;
			break;

		case IF_SWTWORD:	// switch table entry - actually a label ID
			if (m[current].what != IF_SWTWORD)
			{
				current += 1;
				cad += flushdata(output, 1, cad);	// in case there's lurking code
			}
			bufptr = 0;
			putbyte(6);			// Swtab seg
			putw16(swtp);
			id = buffer[0] | (buffer[1] << 8);	// target label number
			ptr = findlabel(id);
			value = labels[ptr].address;
			// the snag is that we now need to relocate this code address
			// we therefore output it as a "zero" and plant a fixup to
			// correct it
			putbyte(0);
			putbyte(0);
			writeobjectrecord(output, 0xA0, bufptr, outbuffer);
			bufptr = 0;
			swtp += WORDSIZE;
			// first plant the locat record.  Offset is zero, since we only
			// ever put one word in each record
			bufptr = 0;
			putbyte(0x80 | (1 << 6) | (1 << 2) | (0 >> 8)); // M = 1, loc = 1
			putbyte(0);	// second part of offset
			// now the fixdat field
			putbyte((0 << 4) | 0);	// Frame is a segment (CODE), target is seg and offset
			putbyte(1);			// index for CSEG
			putbyte(1);			// also index for CODE
			putbyte(value & 255);	// low byte of offset
			putbyte(value >> 8);	// high byte of offset
			writeobjectrecord(output, 0x9C, bufptr, outbuffer);
			bufptr = 0;
			break;

		case IF_LINE:	// line number info for the debugger
			value = buffer[0] | (buffer[1] << 8);
			newlineno(value, cad + codeptr);	// include the code bytes up the spout
			break;

		default:
			fprintf(stderr, "Unexpected tag - not handled\n");
			// all other directives don't consume space
			break;
		}
	}
	flushdata(output, 1, cad);
}

// run through the list of external specs, removing those that
// have not actually been used, and mapping the indexes of those
// that remain to a simple zero-based index
void remapspecs()
{
	int i, index;

	index = 1;		// Both pass2 and OMF files use 1-based indexes
	for (i = 1; i <= nspecs; i++)
	{
		if (specs[i].used)
		{
			specs[i].p3index = index;
			index += 1;
		}
		else
			specs[i].p3index = 0;
	}
	// reassign the specs counter so we know how many we will plant
	nspecs = index;
}

void main(int argc, char **argv)
{
	char * inname;
	char * outname;
	FILE * in;
	FILE * out;

	if (argc < 3)
	{
		fprintf(stderr, "Usage:  IMP2 <intermediatefile> <objfile>\n");
		exit(1);
	}
	inname = argv[1];
	outname = argv[2];

	in = fopen(inname, "r");
	if (in == NULL)
	{
		perror("Can't open input file");
		exit(1);
	}
	// in order to get a useful debug output, we try to recreate the input
	// file name by assuming that the intermediate files have the same base
	// name and are in the same directory as the source.
	_fullpath(path_buffer, inname, _MAX_PATH);			// turn it into a full name
	_splitpath(path_buffer, drive, dir, fname, ext);	// extract components
	_makepath(path_buffer, drive, dir, fname, "imp");	// change the extension to IMP

	readpass1(in);
	fclose(in);

	initlabels();
	while (improvejumpsizes())
		initlabels();
	computesizes();

	remapspecs();

	in = fopen(inname, "r");
	out = fopen(outname, "wb");
	if (out == NULL)
	{
		perror("Can't open output file");
		exit(1);
	}
	readwrite(in, out);

	// now plant the trap table
	puttraptable(out);
	// now output the line number records for the debugger
	putlinenumbers(out);

	putbyte(0);
	writeobjectrecord(out, 0x8A, bufptr, outbuffer);	// end of module

	fclose(in);
	fclose(out);
	dataseg = dataseg+constseg+swtabseg;
	bssseg = bssseg+displayseg;
//	fprintf(stderr, "Items %d   Labels %d   Stackfixes %d   Names %d   Lines %d\n",
//		nm, nl, ns, namedp, nlines);
	fprintf(stderr, "Code %d bytes  Data %d bytes  BSS %d bytes  Total size %d bytes\n",
		codeseg, dataseg, bssseg, codeseg+dataseg+bssseg);

}
